"""Declares :class:`Application`."""
import asyncio
import logging

import fastapi
import ioc
from fastapi.responses import JSONResponse
from unimatrix.exceptions import CanonicalException

from .baserequesthandler import BaseRequestHandler
from .exceptions import TryAgain
from .exceptions import UpstreamServiceNotAvailable
from .exceptions import UpstreamConnectionFailure


class Application(fastapi.FastAPI):
    logger: logging.Logger = logging.getLogger('uvicorn')

    def __init__(self, *args, **kwargs):
        self.log_exceptions = kwargs.pop('log_exceptions', False)
        kwargs.setdefault('docs_url', '/ui')
        kwargs.setdefault('redoc_url', '/docs')
        exception_handlers = kwargs.setdefault('exception_handlers', {})
        exception_handlers.update({
            asyncio.TimeoutError: self.canonical_exception,
            CanonicalException: self.canonical_exception,
            ConnectionError: self.canonical_exception,
            TimeoutError: self.canonical_exception,
        })
        super().__init__(*args, **kwargs)

    def add(self, handler: BaseRequestHandler, *args, **kwargs):
        """Add a request handler to the :class:`Application`."""
        handler.add_to_router(self, *args, **kwargs)

    def configure_query_runners(self, config: dict):
        """Configures and injects the query runners."""
        for alias, runner_class in dict.items(config):
            ioc.provide(f'cbra.runner.{alias}', runner_class)

    async def canonical_exception(self, request, exception):
        """Handles a canonical exception to a standard error message format."""
        if isinstance(exception, ConnectionRefusedError):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamServiceNotAvailable(**kwargs),
            )
        elif isinstance(
            exception,
            (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)
        ):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamConnectionFailure(**kwargs),
            )
        #elif isinstance(exception, UnsatisfiedDependency):
        #    return await self.canonical_exception(
        #        request, FeatureNotSupported()
        #    )
        elif isinstance(exception, (asyncio.TimeoutError, TimeoutError)):
            #count = int(request.headers.get('X-Retry') or 0)
            return await self.canonical_exception(request, TryAgain(30))
        elif isinstance(exception, CanonicalException):
            status_code = getattr(exception, 'http_status_code', None) or 500
            if self.log_exceptions:
                self.logger.error(
                    "%s (%s): %s",
                    exception.code,
                    exception.http_status_code,
                    exception.message
                )
                exception.log(self.logger.exception)
            response = JSONResponse(
                status_code=status_code,
                content=exception.as_dict(),
                headers=exception.http_headers
            )
            return response
        else:
            raise NotImplementedError
