"""Declares common exceptions classes."""
from unimatrix.exceptions import CanonicalException


class Forbidden(CanonicalException):
    """Raised when the request is unconditionally rejected."""
    code = "FORBIDDEN"
    http_status_code = 403
    message = "The request was unconditionally rejected. That's all we know."


class WebhookMessageRejected(CanonicalException):
    """Raised when an incoming webhook message is rejected."""
    code: str = 'MESSAGE_REJECTED'
    http_status_code: int = 403
    message: str = "The server rejected the message."


class AuthenticationRequired(CanonicalException):
    """Raised when an operation required authenticated requests."""
    code = 'AUTHENTICATION_REQUIRED'
    http_status_code = 401
    message = "The operation requires authenticated requests."
    detail = (
        "An operation against a protected resource was attempted and it does "
        "not allow interactions with unauthenticated requests."
    )


class NotAuthorized(CanonicalException):
    """Raised when a request is not authorized to perform a given operation."""
    code = 'UNAUTHORIZED'
    http_status_code = 401
    message = "This request attempts an unauthorized action."
    detail = (
        "The credentials provided with the request did not resolve to a "
        "principal that was authorized to perform the action."
    )


class BearerAuthenticationRequired(AuthenticationRequired):
    hint = "Provide valid credentials using the Authorization header."


class UpstreamServiceNotAvailable(CanonicalException):
    """Raised when the application is not able to establish
    a (network) connection to an upstream service.
    """
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again in 600 seconds."


class UpstreamConnectionFailure(UpstreamServiceNotAvailable):
    """Raised when an upstream service listens at the configured address
    and port, but there are issued in establish the connection according
    to the agreed protocol. Such errors may occur when, for example, the
    upstream service is booting and has bound to its address and port, but
    is not yet ready to serve.
    """
    hint = "Try again in 10 seconds."


class TrustIssues(CanonicalException):
    http_status_code = 403
    code = "TRUST_ISSUES"
    message = (
        "The credential attached to the request was issued by an unkown "
        "authority."
    )
    hint = (
        "Verify that the credential (i.e. bearer token or X.509 certificate) "
        "was issued by an authority that is trusted by this system."
    )


class TryAgain(CanonicalException):
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again in 600 seconds."

    def __init__(self, retry_after): # pragma: no cover
        super().__init__()


class UnsupportedMediaType(CanonicalException):
    http_status_code = 415
    code = "UNSUPPORTED_MEDIA_TYPE"
    message = "The server does not understand the request media type."

    @classmethod
    def fromparsers(cls, parsers: list):
        return cls([x.media_type for x in parsers])

    def __init__(self, accepts: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.accepts = accepts

    def get_http_headers(self) -> dict:
        return {
            **super().get_http_headers(),
            'Accept': str.join(', ', self.accepts)
        }


class ParseError(CanonicalException):
    http_status_code = 400
    code = "MALFORMED_BODY"
    message = (
        "The server understands the media type but the request body "
        "is malformed."
    )


class Unserializable(ValueError):
    pass
