# pylint: skip-file
import base64
import datetime

import aorta
import google.auth
import google.auth.transport.requests
import pytest
from google.auth.exceptions import DefaultCredentialsError
from unimatrix.ext.kms import JSONWebToken
from httpx import AsyncClient

from cbra.lib.auth import NullPrincipal
from cbra.application import Application
from cbra.ext.google.models import MessagePublished
from ..eventarcreceiver import EventarcReceiver
from ..serviceendpoint import ServiceEndpoint


class TestCommand(aorta.Command):
    foo: int


class TestCommandHandler(aorta.CommandHandler):
    pass


class TestEvent(aorta.Event):
    foo: int


class TestEventListener(aorta.EventListener):
    pass



class ProtectedEndpoint(ServiceEndpoint):
    name = 'protected'
    path_parameter = 'resource_id'

    async def list(self):
        pass


class InvalidServiceAccount(ServiceEndpoint):
    name = 'invalid-service-account'
    path_parameter = 'resource_id'

    async def list(self):
        pass


class InvalidAudience(ServiceEndpoint):
    name = 'invalid-audience'
    path_parameter = 'resource_id'

    async def list(self):
        pass


@pytest.fixture(scope='session')
def id_token():
    try:
        credentials, project = google.auth.default()
    except DefaultCredentialsError:
        pytest.skip("Unable to authenticate Google.")
    else:
        request = google.auth.transport.requests.Request()
        credentials.refresh(request)
    return credentials.id_token


@pytest.fixture(scope='session')
def jwt(id_token):
    return JSONWebToken.decode(id_token)


@pytest.fixture(scope='session')
def app(jwt):
    ProtectedEndpoint.members = InvalidAudience.members = [jwt.get('email')]
    ProtectedEndpoint.audience = InvalidServiceAccount.audience = jwt.aud
    app = Application()
    ProtectedEndpoint.add_to_router(app)
    InvalidAudience.add_to_router(app)
    InvalidServiceAccount.add_to_router(app)
    app.add(
        EventarcReceiver(
            members=["foo@bar.baz"],
            principal_class=lambda *a, **k: JSONWebToken.new(
                sub="foo",
                email="foo@bar.baz"
            ),
        ),
        path='/eventarc'
    )
    app.add(
        EventarcReceiver(
            members={"baz@bar.foo"},
            principal=JSONWebToken.new(
                sub="foo",
                email="foo@bar.baz"
            ),
            principal_class=lambda *a, **k: JSONWebToken.new(
                sub="foo",
                email="foo@bar.baz"
            ),
        ),
        path='/eventarc/wrong-email'
    )
    app.add(
        EventarcReceiver(
            principal=JSONWebToken.new(
                sub="foo",
                email="foo@bar.baz"
            ),
            principal_class=lambda *a, **k: JSONWebToken.new(
                sub="foo",
                email="foo@bar.baz"
            ),
        ),
        path='/eventarc/no-members'
    )
    app.add(
        EventarcReceiver(
            principal_class=NullPrincipal
        ),
        path='/eventarc/forbidden'
    )
    return app


@pytest.fixture(scope='session')
def _aorta_register():
    aorta.register(TestCommand, TestCommandHandler)
    aorta.register(TestEvent, TestEventListener)


@pytest.fixture(scope='session')
async def client(app, id_token, jwt):
    headers = {
        'Authorization': f"Bearer {id_token}",
        'Host': list(jwt.aud)[0],
        'Content-Type': "application/json"
    }
    async with AsyncClient(app=app, headers=headers, base_url="http://") as client:
        yield client


@pytest.fixture
def command():
    return TestCommand(foo=1)


@pytest.fixture
def command_b64(command):
    return bytes.decode(base64.b64encode(bytes(command.as_message())), 'ascii')


@pytest.fixture
def command_message(command_b64):
    return MessagePublished(**{
        'subscription': "projects/myproject/subscriptions/mysubscription",
        'message': {
            'messageId': 'id',
            'publishTime': str(datetime.datetime.now()),
            'data': command_b64,
            'attributes': {}
        }
    }).json(by_alias=True)


@pytest.fixture
def command_handler(command):
    return TestCommandHandler(command.as_message())


@pytest.fixture
def event():
    return TestEvent(foo=1)


@pytest.fixture
def event_b64(event):
    return bytes.decode(base64.b64encode(bytes(event.as_message())), 'ascii')


@pytest.fixture
def event_message(event_b64):
    return MessagePublished(**{
        'subscription': "projects/myproject/subscriptions/mysubscription",
        'message': {
            'messageId': 'id',
            'publishTime': str(datetime.datetime.now()),
            'data': event_b64,
            'attributes': {}
        }
    }).json(by_alias=True)


@pytest.fixture
def event_listener(event):
    return TestEventListener(event.as_message())
