"""Declates :class:`NamedQuery`."""
import inspect

import fastapi
import ioc
import pydantic

from .dependency import Dependency
from .dependencywrapper import DependencyWrapper
from .injected import Injected
from .namedquerymetaclass import NamedQueryMetaclass
from .utils import PositionalArgument
from .queryrunner import QueryRunner


class NamedQuery(Dependency, metaclass=NamedQueryMetaclass):
    """A named query that is executed by a dynamically injected
    implementation of :class:`cbra.QueryRunner`.

    The `runner` arguments point to a dependency that is injected at runtime
    under the given key. It is expected to implement :class:`cbra.QueryRunner`
    interface. The default value is ``'QueryRunner'``.
    """
    __abstract__: bool = True
    __module__: str = 'cbra'

    def __init__(self, using: str = 'default'):
        self.using = using
        self.resolve = DependencyWrapper(self.run, [
            PositionalArgument(
                name='runner',
                default=Injected(f'cbra.runner.{self.using}', invoke=True),
                annotation=QueryRunner
            )
        ] + list(inspect.signature(self.build).parameters.values()))
        super().__init__(use_cache=True)

    async def run(self, runner: QueryRunner, *args, **kwargs):
        """Invoke :meth:`build()` with the given positional and keyword
        arguments, and execute the query using `runner`.
        """
        return await runner.execute(self.build(*args, **kwargs))

    def build(self) -> pydantic.BaseModel:
        """Parse the query parameters from the request and execute the query
        using the configured runner.
        """
        raise NotImplementedError
