"""Declares :class:`QueryRunner`."""
import contextlib

import pydantic


class QueryRunner:
    """The base class for all query runners."""
    __module__: str = 'cbra'

    async def connect(self):
        """Setup a connection to the persistent storage system or otherwise
        configure the :class:`QueryRunner` to begin executing queries.
        """
        pass

    async def execute(self, query: pydantic.BaseModel):
        """Executes the given `query`. Subclasses must override this method."""
        raise NotImplementedError(
            f"Unable to run {type(query).__name__}, "
            f"{type(self).__name__}.execute(query) is not implemented."
        )

    async def close(self):
        """Close the connection(s) with the persistent storage system(s) and
        release all resources.
        """
        pass

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args, **kwargs):
        # Since the QueryRunner never does mutations, no special exception
        # handling is needed on exit.
        await self.close()
