"""Declares :class:`RequestHandler`."""
import asyncio
import typing
import inspect

import fastapi
import fastapi.responses

from .bodyconsumer import BodyConsumer
from .exceptions import CanonicalException


class RequestHandler(BodyConsumer):
    """Handles a HTTP request to an endpoint, for a specific request method."""
    _is_coroutine = asyncio.coroutines._is_coroutine

    @property
    def __doc__(self) -> str:
        return self.handle.__doc__

    @property
    def __signature__(self) -> inspect.Signature:
        return inspect.Signature(self.get_signature_parameters())

    def __init__(self,
        method: str,
        handle: typing.Callable,
        is_detail: bool
    ):
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec
        self.handle = handle
        self.is_detail = is_detail
        self.method = method
        self.path = None
        if self.handle is not None:
            # Remove the `self` argument from the signature because it confuses
            # FastAPI (it sees it as a query parameter).
            sig = inspect.signature(self.handle)
            self.handle.__signature__ = sig.replace(
                parameters=[
                    x for x in sig.parameters.values()
                    if x.name != 'self'
                ]
            )

    def add_to_router(self,
        app: fastapi.FastAPI,
        cls: type,
        base_path: str
    ) -> None:
        """Add the :class:`RequestHandler` to a FastAPI ASGI application.

        Args:
            app (:class:`fastapi.FastAPI`): the application to add the resource
                to.
            cls (:class:`cbra.Resource`): the resource class.
            base_path (str): the base path; must start with a slash and not
                end with a slash. Should not contain any path parameters.

        Returns:
            None
        """
        self.resource_class = cls
        self.path = self.get_mount_path(base_path)
        app.add_api_route(
            path=self.path,
            endpoint=self,
            methods=[self.method]
        )

    def get_signature_parameters(self) -> list:
        """Get the list of :class:`inspect.Parameter` instances specifying
        the callables' signature.
        """
        params = [
            inspect.Parameter(
                name='request',
                kind=inspect.Parameter.POSITIONAL_ONLY,
                annotation=fastapi.Request
            ),
            inspect.Parameter(
                name='response',
                kind=inspect.Parameter.POSITIONAL_ONLY,
                annotation=fastapi.responses.Response
            ),
        ] + self.get_parent_path_parameters()
        if self.is_detail:
            params.append(self.resource_class.sig_path_parameter)

        # Add dependencies.
        params.extend([
            inspect.Parameter(
                name='principal',
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=self.resource_class.principal,
                annotation=dict
            )
        ])
        return params

    def get_parent_path_parameters(self) -> list:
        """Return a list of :class:`inspect.Signature` instance representing
        the path parameters of a subresources' parents.
        """
        params = []
        cls = self.resource_class.parent
        while cls is not None:
            params.append(cls.sig_path_parameter)
            cls = cls.parent
        return list(reversed(params))

    def get_method_handler(self, resource) -> typing.Callable:
        """Return the handle function for the current request method."""
        return self.handle

    def get_mount_path(self, base_path: str) -> str:
        """Return the mount path of the :class:`RequestHandler`."""
        path = base_path
        if self.is_detail:
            path = f'{base_path}/{{{self.resource_class.path_parameter}}}'
        return path

    async def __call__(self,
        request: fastapi.Request,
        response: fastapi.responses.Response,
        principal: dict,
        *args, **kwargs
    ):
        """Resolve dependencies and invoke the appropriate method handler on
        the resource class specified in the constructor.
        """
        assert self.path is not None # nosec
        resource = self.resource_class(
            path=self.path,
            principal=principal,
            request=request,
            response=response,
            args=args,
            kwargs=kwargs
        )
        try:
            return await resource.dispatch(
                handle=self.get_method_handler(resource)
            )
        except CanonicalException as exc:
            resource.response.status_code = exc.http_status_code
            resource.response.headers.update(exc.http_headers)
            return exc.as_dict()
