"""Declares :class:`Resource`."""
import inspect
import logging
import typing

import fastapi
import fastapi.params
from fastapi import Request
from fastapi.responses import Response
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from .application import Application
from .bodyconsumer import BodyConsumer
from .exceptions import AuthenticationRequired
from .ext.auth import CurrentPrincipal
from .resourcemetaclass import ResourceMetaclass
from .requesthandler import RequestHandler


class Resource(BodyConsumer, metaclass=ResourceMetaclass):
    """Represents a resource in a web API."""
    __abstract__ = True
    logger: logging.Logger = logging.getLogger('uvicorn')
    parent: ResourceMetaclass = None

    #: The name of the resource.
    name: str

    #: The name of the resource identifier as a path parameter.
    path_parameter: str

    #: Indicates if authentication is required.
    require_authentication: bool = True

    #: The dependency used to resolve the current principal.
    principal = CurrentPrincipal

    #: The subresources for this resource.
    subresources: list = []

    #: Additional dependencies specified for this class.
    dependencies: typing.List[fastapi.params.Depends] = []

    @classmethod
    def add_to_router(cls,
        app: fastapi.FastAPI,
        base_path: str = '',
        path: str = None
    ) -> None:
        """Add the :class:`Resource` to a FastAPI ASGI application.

        Args:
            app (:class:`fastapi.FastAPI`): the application to add the resource
                to.
            base_path (str): the base path; must start with a slash and not
                end with a slash. Should not contain any path parameters.
            path (str): if not ``None``, use an explicit path instead of the
                :attr:`name` attribute.

        Returns:
            None
        """
        base_path = f'{base_path}/{str.lstrip(path or cls.name, "/")}'
        for _, handler in\
        inspect.getmembers(cls, lambda attr: isinstance(attr, RequestHandler)):
            handler.add_to_router(app, cls, base_path)
        for sub in cls.subresources:
            sub.add_to_router(
                app=app,
                base_path=f'{base_path}/{{{cls.path_parameter}}}'
            )

    @classmethod
    def run_simple(cls, *args, **kwargs): # pragma: no cover
        # TODO: this is for demo purposes only
        import cbra

        app = Application(docs_url='/ui')
        cls.add_to_router(app)
        cbra.run(app, *args, **kwargs)

    def __init__(self,
        request: Request,
        response: Response,
        path: str,
        principal: dict,
        args: tuple,
        kwargs: dict
    ):
        self.args = args
        self.kwargs = kwargs
        self.path = path
        self.principal = principal
        self.request = request
        self.response = response

    def get_dependants(self):
        return []

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return bool(self.principal)

    async def authorize(self) -> bool:
        """Hook to perform authorization checks for the request. Must raise
        an appropriate exception if the request is not authorized. This
        method may specify dependencies.
        """
        pass

    async def dispatch(self, handle: typing.Callable):
        if self.require_authentication and not self.is_authenticated():
            raise AuthenticationRequired
        dependant = get_dependant(path=self.path, call=handle)
        dependant.dependencies.extend(self.__get_dependants())
        values, errors, tasks, response, _ =  await solve_dependencies(
            request=self.request,
            dependant=dependant,
            body=await self.get_body(),
            dependency_overrides_provider=None
        )
        if response.status_code is not None: # pragma: no cover
            # TODO
            raise NotImplementedError
        if tasks is not None: # pragma: no cover
            # TODO
            raise NotImplementedError
        if errors: # pragma: no cover
            # TODO
            raise NotImplementedError(errors)
        return await dependant.call(self, **values)

    def __get_dependants(self):
        dependants = self.get_dependants()
        dependants.extend([
            self.authorize
        ])
        return [
            get_parameterless_sub_dependant(
                depends=fastapi.Depends(func),
                path=self.path
            ) for func in dependants
        ]
