"""Declares :class:`ResourceMetaclass`."""
import abc
import copy
import inspect

from .requesthandler import RequestHandler
from .optionshandler import OptionsHandler


ACTIONS = [
    ('GET', 'list', False),
    ('POST', 'create', False),
    ('DELETE', 'purge', False),
    ('HEAD', 'exists', True),
    ('GET', 'retrieve', True),
    ('POST', 'receive', True),
    ('PATCH', 'update', True),
    ('PUT', 'replace', True),
    ('DELETE', 'destroy', True),
]
DETAIL_ACTIONS = {'receive', 'retrieve', 'update', 'delete', 'replace'}


class ResourceMetaclass(abc.ABCMeta):

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        if attrs.pop('__abstract__', False):
            return super_new(cls, name, bases, attrs)

        annotations = attrs.get('__annotations__') or {}
        if not attrs.get('name'): # pragma: no cover
            raise ValueError(f"Define {name}.name")
        if cls.has_detail_methods(attrs) or cls.has_subresources(attrs):
            # If the class specified detail methods or subresources, then make
            # the path_parameter mandatory.
            if 'path_parameter' not in attrs: # pragma: no cover
                raise ValueError(f"Define {name}.path_parameter")
            else:
                attrs['sig_path_parameter'] = inspect.Parameter(
                    name=attrs.get('path_parameter'),
                    kind=inspect.Parameter.POSITIONAL_ONLY,
                    annotation=annotations.get('path_parameter') or str
                )

        subresources = attrs.pop('subresources', [])
        new_class = super_new(cls, name, bases, attrs)

        # Set the new class as the parent for all subresources. Since
        # subresources may be used by multiple resources, copy the class.
        for i, subresource in enumerate(subresources):
            subresources[i] = copy.deepcopy(subresource)
            subresources[i].parent = new_class
        setattr(new_class, 'subresources', subresources)

        # Find all actions and wrap the methods in a RequestHandler
        # instance.
        detail_methods = []
        list_methods = []
        for method, action, is_detail in ACTIONS:
            if action not in attrs:
                continue
            handle = getattr(new_class, action)
            if isinstance(handle, RequestHandler):
                # Nothing to do for manually crafted RequestHandler
                # instances.
                continue
            handler_class = getattr(handle, 'handler_class', RequestHandler)
            func = handler_class(
                method=method,
                handle=handle,
                is_detail=is_detail
            )
            setattr(new_class, action, func)
            detail_methods.append(method)\
                if is_detail else\
                list_methods.append(method)

        if detail_methods:
            setattr(new_class, '_options_detail', OptionsHandler(
                allowed_methods=detail_methods,
                method='OPTIONS',
                handle=None,
                is_detail=True
            ))
        if list_methods:
            setattr(new_class, '_options_list', OptionsHandler(
                allowed_methods=list_methods,
                method='OPTIONS',
                handle=None,
                is_detail=False
            ))

        return new_class

    @staticmethod
    def has_detail_methods(attrs: dict) -> bool:
        return bool(set(attrs.keys()) & DETAIL_ACTIONS)

    @staticmethod
    def has_subresources(attrs: dict) -> bool: # pragma: no cover
        return bool(attrs.get('subresources'))
